@@ -8,7 +8,7 @@ const ipv6normalize = require('ipv6-normalize');
 const sasl = require('./sasl');
 const crypto = require('crypto');
 const os = require('os');
-const punycode = require('punycode');
+const punycode = require('punycode/');
 const EventEmitter = require('events');
 const base32 = require('base32.js');
 
@@ -160,6 +160,21 @@ class SMTPConnection extends EventEmitter {
 
             this._resetSession();
 
+            let onSecureIfNeeded = next => {
+                if (!this.session.secure) {
+                    // no TLS
+                    return next();
+                }
+
+                this.session.servername = this._socket.servername;
+                this._server.onSecure(this._socket, this.session, err => {
+                    if (err) {
+                        return this._onError(err);
+                    }
+                    next();
+                });
+            };
+
             this._server.onConnect(this.session, err => {
                 this._server.logger.info(
                     {
@@ -177,20 +192,25 @@ class SMTPConnection extends EventEmitter {
                     return this.close();
                 }
 
-                this._ready = true; // Start accepting data from input
+                onSecureIfNeeded(() => {
+                    this._ready = true; // Start accepting data from input
 
-                if (!this._server.options.useXClient && !this._server.options.useXForward) {
-                    this.emitConnection();
-                }
+                    if (!this._server.options.useXClient && !this._server.options.useXForward) {
+                        this.emitConnection();
+                    }
 
-                this.send(
-                    220,
-                    this.name + ' ' + (this._server.options.lmtp ? 'LMTP' : 'ESMTP') + (this._server.options.banner ? ' ' + this._server.options.banner : '')
-                );
+                    this.send(
+                        220,
+                        this.name +
+                            ' ' +
+                            (this._server.options.lmtp ? 'LMTP' : 'ESMTP') +
+                            (this._server.options.banner ? ' ' + this._server.options.banner : '')
+                    );
 
-                if (typeof next === 'function') {
-                    next();
-                }
+                    if (typeof next === 'function') {
+                        next();
+                    }
+                });
             });
         };
 
@@ -476,7 +496,13 @@ class SMTPConnection extends EventEmitter {
         }
 
         // block users that try to fiddle around without logging in
-        if (!this.session.user && this._isSupported('AUTH') && commandName !== 'AUTH' && this._maxAllowedUnauthenticatedCommands !== false) {
+        if (
+            !this.session.user &&
+            this._isSupported('AUTH') &&
+            !this._server.options.authOptional &&
+            commandName !== 'AUTH' &&
+            this._maxAllowedUnauthenticatedCommands !== false
+        ) {
             this._unauthenticatedCommands++;
             if (this._unauthenticatedCommands >= this._maxAllowedUnauthenticatedCommands) {
                 return this.send(421, 'Error: too many unauthenticated commands');
@@ -490,7 +516,10 @@ class SMTPConnection extends EventEmitter {
 
         // Check if authentication is required
         if (!this.session.user && this._isSupported('AUTH') && ['MAIL', 'RCPT', 'DATA'].includes(commandName) && !this._server.options.authOptional) {
-            this.send(530, 'Error: authentication Required');
+            this.send(
+                530,
+                typeof this._server.options.authRequiredMessage === 'string' ? this._server.options.authRequiredMessage : 'Error: authentication Required'
+            );
             return setImmediate(callback);
         }
 
@@ -1423,6 +1452,7 @@ class SMTPConnection extends EventEmitter {
             this._upgrading = false;
 
             this.session.tlsOptions = this.tlsOptions = this._socket.getCipher();
+            this.session.servername = this._socket.servername;
             let cipher = this.session.tlsOptions && this.session.tlsOptions.name;
             this._server.logger.info(
                 {
@@ -1434,10 +1464,15 @@ class SMTPConnection extends EventEmitter {
                 'Connection upgraded to TLS using ',
                 cipher || 'N/A'
             );
-            this._socket.pipe(this._parser);
-            if (typeof secureCallback === 'function') {
-                secureCallback();
-            }
+            this._server.onSecure(this._socket, this.session, err => {
+                if (err) {
+                    return this._onError(err);
+                }
+                this._socket.pipe(this._parser);
+                if (typeof secureCallback === 'function') {
+                    secureCallback();
+                }
+            });
         });
     }
 }
